#define NDEBUG
#define _SCL_SECURE_NO_WARNINGS

#ifdef __BORLANDC__
#pragma warn-stl // disable a wrong warning in BCC
#endif

#include "algorithm.h"
#include "micro_timer.h"
#include "rand_bool.h"

#include <algorithm>
#include <malloc.h>
#include <stdexcept>
#include <vector>

#define array_alloca(type, size) \
    (static_cast<type *>(alloca((size) * sizeof(type))))

using namespace std;

typedef size_t in_val_count_t;

static span span_distance(const span &ref_entry1, const span &ref_entry2) {
    return
        ref_entry1 >= ref_entry2
        ? ref_entry1 - ref_entry2
        : ref_entry2 - ref_entry1;
}

static void sort_entries(
const vector<span> &ref_entry_vector,
span entry_array[])
{
    vector<span>::const_iterator begin = ref_entry_vector.begin();
    entry_count_t entry_array_length = ref_entry_vector.size() - 1;
    partial_sort_copy(
        ++begin,
        ref_entry_vector.end(),
        entry_array,
        entry_array + entry_array_length
        );
}

static void call_node_listener(
const vector<span> &ref_entry_vector,
solve_progress_listener *p_listener,
micro_timer &timer)
{
    micro_timer pause_timer;
    span *entry_array;
    {
        entry_count_t entry_array_length = ref_entry_vector.size() - 1;
        entry_array = array_alloca(span, entry_array_length);
    }
    sort_entries(ref_entry_vector, entry_array);
    p_listener->node_visited(entry_array, ref_entry_vector.size());
    timer -= pause_timer;
}

static int
get_new_entries(
const vector<span> &ref_entry_vector,
const span &ref_largest,
span new_entry_array[2])
{
    // add primary new entry moving forth from the start
    new_entry_array[0] = ref_largest;
    // Calculate symmetrical new entry moving back from the end except when the
    // number of calculated entries is less than 3: the first entry is always 0;
    // the second entry is always the largest one; the third entry produces a
    // partial solution that is all the way symmetrical to the solution produced
    // by adding the primary entry.
    if (ref_entry_vector.size() > 2) {
        span new_entry;
        {
            vector<span>::const_iterator begin = ref_entry_vector.begin();
            new_entry = *++begin - ref_largest;
        }
        // add second new entry if it's different from the first one
        if (new_entry != ref_largest) {
            new_entry_array[1] = new_entry;
            // shuffle randomly
            if (rand_bool()) swap(new_entry_array[0], new_entry_array[1]);
            return 1;
        }
    }
    return 0;
}

static void
backtracking_search(
const span in_val_array[],
vector<bool> exclusion_vector,
vector<span> &ref_entry_vector,
unsigned int max_solution_count,
unsigned int max_visited_node_count,
const span &ref_entry,
solve_progress_listener *p_listener,
solution_set &ref_best_set,
unsigned int &ref_visited_node_count,
micro_timer &ref_timer)
{
    for (
        vector<span>::iterator iterator = ref_entry_vector.begin();
        iterator != ref_entry_vector.end();
        ++iterator)
    {
        span gap = span_distance(ref_entry, *iterator);
        for (
            in_val_count_t in_val_index =
            lower_bound(
            in_val_array,
            in_val_array + exclusion_vector.size(),
            gap) -
            in_val_array;;
            ++in_val_index)
        {
            if (
                in_val_index >= exclusion_vector.size() ||
                in_val_array[in_val_index] != gap)
                return;
            if (!exclusion_vector[in_val_index]) {
                exclusion_vector[in_val_index] = true;
                break;
            }
        }
    }
    ref_entry_vector.push_back(ref_entry); // set new entry point
    ++ref_visited_node_count;
    if (p_listener != NULL)
        call_node_listener(ref_entry_vector, p_listener, ref_timer);
    // if all entry points are set
    if (ref_entry_vector.size() == ref_best_set.get_entry_count()) {
        // primary solution
        solution best_1(ref_best_set.get_entry_count() - 1);
        sort_entries(ref_entry_vector, best_1);
        if (ref_best_set.get_entry_count() > 2) {
            // symmetric solution
            solution best_2(ref_best_set.get_entry_count() - 1);
            {
                entry_count_t index_1 = ref_best_set.get_entry_count() - 2;
                entry_count_t index_2 = 0;
                const span &ref_last = best_1[index_1];
                while (index_1 > 0) {
                    const span &ref_symmetric_entry = best_1[--index_1];
                    span entry = ref_last - ref_symmetric_entry;
                    best_2[index_2++] = entry;
                }
                best_2[index_2] = ref_last;
            }
            if (ref_best_set.size() + 1 >= max_solution_count && rand_bool())
                solution::swap(best_1, best_2);
            if (ref_best_set.size() < max_solution_count)
                ref_best_set.insert(best_2);
        }
        ref_best_set.insert(best_1);
    }
    else if (ref_visited_node_count < max_visited_node_count) {
        // create set of new entries
        int child_index;
        span new_entry_array[2];
        {
            in_val_count_t largest_index = exclusion_vector.size();
            while (exclusion_vector[--largest_index]);
            const span &ref_largest = in_val_array[largest_index];
            child_index =
                get_new_entries(ref_entry_vector, ref_largest, new_entry_array);
        }
        do {
            span &ref_new_entry = new_entry_array[child_index];
            backtracking_search(
                in_val_array,
                vector<bool>(exclusion_vector),
                ref_entry_vector,
                max_solution_count,
                max_visited_node_count,
                ref_new_entry,
                p_listener,
                ref_best_set,
                ref_visited_node_count,
                ref_timer
                );
            --child_index;
        }
        while (
            child_index >= 0 &&
            ref_best_set.size() < max_solution_count &&
            ref_visited_node_count < max_visited_node_count
            );
    }
    ref_entry_vector.pop_back(); // revert entry vector
}

static entry_count_t get_entry_count(in_val_count_t in_val_count) {
    entry_count_t entry_count = 0;
    in_val_count_t in_val_left = in_val_count;
    do {
        ++entry_count;
        if (in_val_left == 0) return entry_count;
        in_val_left -= entry_count;
    }
    while (in_val_left < in_val_count);
    return 0;
}

unsigned int
solve(
const multiset<span> &ref_in_val_multiset,
unsigned int max_solution_count,
unsigned int max_visited_node_count,
solve_progress_listener *p_listener,
solution_set &ref_best_set,
unsigned long long &ref_elapsed_time) {
    micro_timer timer;
    in_val_count_t size = ref_in_val_multiset.size();
    if (size > 0 && *ref_in_val_multiset.begin() == 0)
        throw invalid_argument("all input values must be positive integers");
    entry_count_t entry_count = get_entry_count(size);
    if (entry_count == 0)
        throw invalid_argument("invalid number of distances");
    span *in_val_array = array_alloca(span, size);
    copy(
        ref_in_val_multiset.begin(),
        ref_in_val_multiset.end(),
        in_val_array
        );
    ref_best_set.set_entry_count(entry_count);
    unsigned int visited_node_count = 0;
    if (max_solution_count > 0 && max_visited_node_count > 0) {
        vector<bool> exclusion_vector;
        exclusion_vector.resize(size);
        vector<span> entry_vector;
        backtracking_search(
            in_val_array,
            exclusion_vector,
            entry_vector,
            max_solution_count,
            max_visited_node_count,
            0,
            p_listener,
            ref_best_set,
            visited_node_count,
            timer
            );
    }
    ref_elapsed_time = timer;
    return visited_node_count;
}

solution_comparator::solution_comparator(const entry_count_t &ref_entry_count):
_ref_entry_count(ref_entry_count) {
}

bool
solution_comparator::operator()(const solution &arg1, const solution &arg2)
const {
    assert(_ref_entry_count > 0);
    return
        lexicographical_compare(
        arg1.begin(),
        arg1.end(_ref_entry_count - 1),
        arg2.begin(),
        arg2.end(_ref_entry_count - 1)
        );
}

solution_set::solution_set():
set<solution, solution_comparator>(solution_comparator(_entry_count)),
_entry_count(0) {
}

entry_count_t solution_set::get_entry_count() const {
    return _entry_count;
}

void solution_set::set_entry_count(entry_count_t entry_count) {
    _entry_count = entry_count;
}
