#define WIN32_LEAN_AND_MEAN

#include "micro_timer.h"

#include <stdexcept>
#include <windows.h>

using namespace std;

static void validate_api_call(BOOL result) {
    if (!result) throw runtime_error("timer not available");
}

static void get_counter(long long &ref_counter) {
    validate_api_call(
        QueryPerformanceCounter(
        reinterpret_cast<LARGE_INTEGER *>(&ref_counter)
        ));
}

static long long get_frequency() {
    LARGE_INTEGER freq;
    validate_api_call(QueryPerformanceFrequency(&freq));
    return freq.QuadPart;
}

micro_timer::micro_timer() {
    get_counter(_start);
}

micro_timer::operator unsigned long long() const {
    return
        static_cast<unsigned long long>
        (get_elapsed_time() / (get_frequency() * .000001) + .5);
}

micro_timer &micro_timer::operator +=(const micro_timer &arg) {
    _start -= arg.get_elapsed_time();
    return *this;
}

micro_timer &micro_timer::operator -=(const micro_timer &arg) {
    _start += arg.get_elapsed_time();
    return *this;
}

long long micro_timer::get_elapsed_time() const {
    long long counter;
    get_counter(counter);
    return counter - _start;
}
